#!/usr/local/bin/perl -w
# File:     sample_postfilter_backoff.pl
# Created:  24-Jul-1999

# COPYRIGHT 1999
# bCandid Corporation.
# All Rights Reserved.
#
# software@bcandid.com
# 303-545-5550

# This sample filter for posts implements a "sliding window" scheme
# preventing an authenticated user from posting more than $postlimit
# articles during any $interval (in seconds) period. Posts from
# unauthenticated users are rejected (since this would totally
# circumvent the back-off scheme).

# --- Begin user-configurable constants ---

# Allow $postlimit POSTs every $interval seconds.
my $interval = 20;
my $postlimit = 5;

# --- End user-configurable constants ---

# An associative array mapping a user's name to a list of the times at which
# the user's last $postlimit POSTs occurred (sorted ascending).
my %posts = ();

# $delay = access_delay($client, $time, $interval, $postlimit)
#
# Returns $delay, the minimum amount of time to delay so that the user
# ($client) will not have been able to post more than $postlimit messages
# in a "sliding window" of $interval seconds. $time is the current time.
# $time most be nondecreasing across sequential calls.

sub access_delay
{
    my ($client, $time, $interval, $postlimit) = @_;

    # Get the appropriate access-time list.
    my $times = $posts{$client};
    if (!defined($times)) {
	# Never seen $client before - add to the assoc. array.
	$times = $posts{$client} = [];
    }

    # Forget all posts which took place at least $interval ago.
    shift @$times while (@$times > 0 && $time - $$times[0] >= $interval);

    if (@$times == $postlimit) {
	# No more posts permitted for a while! Remove the first (oldest)
	# time, but back off until $interval past that time. Record the
	# post as actually occurring after the back-off period.
	push(@$times, $$times[0] + $interval);
	my $delay = $$times[0] + $interval - $time;
	shift @$times;
	return $delay;
    } else {
	# Record the post.
	push(@$times, $time);
	return 0;
    }
}

# Flush STDOUT
$| = 1;
$previous_line = "\r\n";

# Enter a simple state machine
while (defined($line = <>)) {
    # If we just completed the article send a response
    if ($previous_line =~ /\r\n$/ && $line eq ".\r\n") {
	if ($username) {
	    # The user authenticated; apply the back-off scheme.

	    my $delay = access_delay($username, time(), $interval, $postlimit);
	    if ($delay == 0) {
		print "235\r\n";
	    } else {
		print "236 $delay\r\n";
	    }
	} else {
	    # Do not accept unauthenticated POSTs!
	    print "435\r\n";
	}

	# Terminate the response
	print ".\r\n";
	$username = $in_header = $in_body = undef;
    }

    # We are fed:
    #   FeedInformationLines\r\n
    #   \r\n
    #   HeaderLines\r\n
    #   \r\n
    #   BodyLines\r\n
    #   .\r\n
    #
    # So, set up simple state to save the right lines
    #

    # If we are in the FeedInformation area, grab Authentication
    #  information for later use
    if (!$in_header && !$in_body) {
	if ($line =~ m/^Username: (.*)\r\n$/) {
	    $username = $1;
	}
    }

    if ($line eq "\r\n") {
	if ($in_header) {
	    $in_body = 1;
	} else {
	    $in_header = 1;
	}
    }

    # Save last line for use in end of transaction detection
    $previous_line = $line;
}
