#!/usr/local/bin/perl
# $Header: /opt/nr/rcs/cmd/scripts/log-report.pl,v 1.19 1999/07/01 23:26:45 rmf Exp $
#
# log-report.pl
# Copyright (c) 1997-1999 bCandid Corporation
#
# Summary:
#
#   This perl script produces some statistics from a Cyclone, Typhoon,
# or Breeze "-log" file.  There are four different modes of operation:
#
# - incoming article rate (-r) (DEFAULT)
# - incoming article source (-s)
# - incoming article distribution by article size (-d)
# - spool usage (-u)
#
# Multiple options can be specified on the same line.  Non-options are
# treated as regular expressions.  If regular expressions are used, only
# lines which match at least one regular expression will be reported on.
#
# It will correctly handle an incomplete last line (caused by using
# tail on a currently expanding log file)
#
# Example usage:
#
# To produce a full summary:
#   $ cat log/log | tools/log-report.pl -r
#   $ cat log/log | tools/log-report.pl -r -w60
#   $ cat log/log | tools/log-report.pl -s
#   $ cat log/log | tools/log-report.pl -s -p
#   $ cat log/log | tools/log-report.pl -ss
#   $ cat log/log | tools/log-report.pl -ss -p
#   $ cat log/log | tools/log-report.pl -d
#   $ cat log/log | tools/log-report.pl -u
#   $ cat log/log | tools/log-report.pl -u -c
#   $ cat log/log | tools/log-report.pl -r -ss -d -u
# To produce a summary on recent incoming articles:
#   $ tail -500 log/log | tools/log-report.pl -r
#   $ tail -500 log/log | tools/log-report.pl -r -w5
#   $ tail -500 log/log | tools/log-report.pl -s
#   $ tail -500 log/log | tools/log-report.pl -s -p
#   $ tail -500 log/log | tools/log-report.pl -ss
#   $ tail -500 log/log | tools/log-report.pl -ss -p
#   $ tail -500 log/log | tools/log-report.pl -d
#   $ tail -500 log/log | tools/log-report.pl -u
#   $ tail -500 log/log | tools/log-report.pl -u -c
# To summarize specific data:
#   $ cat log/log | tools/log-report.pl -r REGEXP
#   $ cat log/log | tools/log-report.pl -d REGEXP REGEXP2
#
# Notes:
#
# Incoming Article Rate:
#   Rates for the first and last seconds will often be "low" due to other
#   articles in the second not being included in the log sample.
#
# Incoming Article Source: (-ss)
#   Specifying -ss will display incoming articles by IncomingHostname
#   and IP address pair; normally incoming articles are displayed by
#   IncomingHostname only
#
# Incoming Article Distribution By Article Size:
#   Modify the @sizes array to your liking for different size groups
#
# Spool Usage:
#   This only works for Typhoon or Breeze log files and assumes that the
#   data is in chronological order.  To make the Lifetime field show the
#   correct data for your server, add @spoolsize entries to be the
#   correct spool size
#
# Spool Compression option: (-c)
#   This option works in conjunction with the Spool Usage mode of
#   operation.  It gives information on a per spool basis about the
#   compression of articles received.
#
# Window size option: (-w)
#   This option examines incoming articles with a base unit time other
#   than a single second.  Used with the -r  option, this specifies that
#   articles will reported on looking at X seconds where X is the argument to
#   -w instead of on a 1 second granularity.  There should be no spaces in
#   between the option and the numerical value
#
# Per second option: (-p)
#   This option, valid with -s or with -r and -w, averages data for the
#   entire period being reported to a base unit of 1 second.  For example,
#   normally the -s options gives the total number of articles and bytes
#   for the entire log period.  With the -p option, the report gives
#   articles/second and bytes/second for each Incoming feed.
#   This makes use of the fact that the times in the log file are stored
#   to microsecond precision.  It only affects the Incoming Article Source
#   option and the Incoming Article Rate w/ Window size
#

# Size of each spool for Spool Usage in Kilobytes (Typhoon/Breeze only)
#   $defaultsize is used for any spool that is not listed below
$defaultsize     = 2000000;
$SPOOLSIZE[1]    = 2000000;
$SPOOLSIZE[2]    = 2000000;
$SPOOLSIZE[3]    = 2000000;
$SPOOLSIZE[4]    = 2000000;
$SPOOLSIZE[5]    = 2000000;
$SPOOLSIZE[6]    = 2000000;
$SPOOLSIZE[7]    = 2000000;
$SPOOLSIZE[8]    = 2000000;
$SPOOLSIZE[9]    = 2000000;
$SPOOLSIZE[10]   = 2000000;
# add more spools above this line

# Article sizes in ascending order, one per line, for the incoming
#   article distribution mode -- do not remove the 0 element

@SIZE = (
  0,
  100,
  200,
  300,
  400,
  500,
  600,
  700,
  800,
  900,
  1000,
  1100,
  1200,
  1300,
  1400,
  1500,
  1600,
  1700,
  1800,
  1900,
  2000,
  2500,
  3000,
  3500,
  4000,
  5000,
  6000,
  8000,
  12000,
  16000,
  32000,
  48000,
  64000,
  128000,
  256000,
  512000,
  1024000,
  1536000,
  2048000
);

# Constants (modes) - Don't modify
$RATE = 1;
$SRC  = 2;
$DIST = 3;
$USE  = 4;
$INC  = 5;
$COMP = 6;

# Default mode (feel free to override)
$defaultmode = $RATE;
$defaultmodevalue = 1;		# For $SRC
$windowsize = 0;		# For $RATE
$persecond = 0;			# For $RATE and $SRC

#######################################
## End of configuration

# Convert the spoolsizes to bytes
@SPOOLSIZE = map( { $_ * 1024 } @SPOOLSIZE);
$defaultsize *= 1024;

# Ensure that @sizes is sorted.
@SIZE = sort { $a <=> $b } (@SIZE);
$SIZEcachesize = 8000;

# Variable initialization
@modelist = ();
$action = 0;
$starttime = 0;
$endtime = 0;
$startstr = "";
$endstr = "";
$timediff = 0;
$timestr = "";

# Number of fields in a correctly formatted -log line
$numfields = 7;
$numfieldstyphoon = 8;
$numfieldstyphoonlg = 9;
$intimefield = 0;
$inhostfield = 1;
$inipfield = 2;
#$inmsgidfield = 3;
$insizefield = 4;
#$incrossfield = 5;
#$indatefield = 6;
$inspoolfield = 7;
#$ingroupfield = 8;

$TOTALicount = 0;
$TOTALibytes = 0;
$TOTALfcount = 0;
$TOTALfbytes = 0;
$TOTALucount = 0;
$TOTALubytes = 0;
$TOTALccount = 0;
$TOTALcbytes = 0;
$TOTALicbytes = 0;
$TOTALdcount = 0;
$TOTALdbytes = 0;
#######################################
## Command line parsing

# Parse command line
ARG: foreach (@ARGV) {
  if (/^-r/) {
    $modelist[$RATE] = 1;
    $action++;
    next ARG;
  }

  if (/^-s/) {
    $modelist[$SRC] = 1;
    $action++;
    if (/^-ss/) {
      $modelist[$SRC] = 2;
    }
    next ARG;
  }

  if (/^-d/) {
    $modelist[$DIST] = 1;
    $action++;

    # initialization
    for $index (0 .. $#SIZE) {
      $DISTcount[$index] = 0;
      $DISTbytes[$index] = 0;
    }

    $index = 0;
    $max = $SIZE[$index];
    for $size (0 .. $SIZEcachesize) {
      GROUP: while ($size > $max) {
	$index++;
	$max = $SIZE[$index];
	if ($max == 0) {
	  last GROUP;
	}
      }
      $SIZEcache[$size] = $index;
    }

    next ARG;
  }

  if (/^-u/) {
    $modelist[$USE] = 1;
    $action++;
    next ARG;
  }

#  if (/^-i/) {
#    $modelist[$INC] = 1;
#    $action++;
#    next ARG;
#  }

  if (/^-c/) {
    $modelist[$USE] = 1;
    $modelist[$COMP] = 1;
    $action++;
    next ARG;
  }

  if (/^-w(\d*)/) {
    $windowsize = $1;
    if ($windowsize eq "") {
      print(STDERR "WARNING: no argument to -w included\n");
      $windowsize = 0;
    }
    next ARG;
  }

  if (/^-p/) {
    $persecond = 1;
    next ARG;
  }

  if (/^-/) {
    print(STDERR "WARNING: unknown option $_ ignored\n");
    next ARG;
  }

  push (@regexps, $_);
}

if ($action == 0) {
  $modelist[$defaultmode] = $defaultmodevalue;
}

#######################################
## Read data

LINE: while (defined($line = <STDIN>)) {

  # Drop unterminated lines
  unless ($line =~ /\n$/) {
    next LINE;
  }

  # Get rid of trailing newline
  chomp($line);

  # Check for regular expression matching
  if ($#regexps != -1) {
    $pass = 0;
    RE: foreach $regexp (@regexps) {
      if ($line =~ /$regexp/) {
	$pass = 1;
	last RE;
      }
    }

    if ($pass == 0) {
      next LINE;
    }
  }

  # Split up the tab split line
  @fields = split(/\t/, $line);

  # First line, special parsing
  if ($starttime == 0) {
    # Count the number of fields
    if ($#fields == ($numfieldstyphoon - 1)) {
      $numfields = $numfieldstyphoon;
    } elsif ($#fields == ($numfieldstyphoonlg - 1)) {
      $numfields = $numfieldstyphoonlg;
    } else {
      # Default number of fields (Cyclone)
      if ($#fields >= 0 && $modelist[$USE] > 0) {
	print(STDERR
	  "ERROR: Spool Usage mode can only be used with Typhoon or Breeze\n");
	exit(1);
      }
    }
  }

  # Check if we've got the right number of fields
  if ($#fields != ($numfields - 1)) {
    print(STDERR "WARNING: wrong number of fields: $line\n");
    next LINE;
  }

# counts and bytes
# initial - whats coming in
# final - whats actually stored to disk
# comp - for compressed articles only
# uncomp - for uncompressed articles only
# deleted - for compressed to 0
  # Save repeatedly used fields
  ($timefield) = ($fields[$intimefield] =~ /^(\d+)\./);

  if ($fields[$insizefield] =~ /^(\d+)\/(\d+)/) {
    $sizeuncomp = $1;
    $sizecomp = $2;
  } else {
    $sizeuncomp = $sizecomp = $fields[$insizefield];
  }

  $TOTALicount ++;
  $TOTALibytes += $sizeuncomp;
  $TOTALfbytes += $sizecomp;
  if ($sizecomp == 0) {
    $TOTALdcount ++;
    $TOTALdbytes += $sizeuncomp;
    $sizemode = 0;
  } else {
    $TOTALfcount ++;
    if ($sizeuncomp == $sizecomp) {
      $TOTALucount ++;
      $TOTALubytes += $sizeuncomp;
      $sizemode = 1;
    } else {
      $TOTALccount ++;
      $TOTALicbytes += $sizeuncomp;
      $TOTALcbytes += $sizecomp;
      $sizemode = -1;
    }
  }

  $TOTALcount = $TOTALicount;
  $TOTALbytes = $TOTALfbytes;

  if ($starttime == 0) {
    $starttime = $timefield;
    $WINDOWend = $starttime + $windowsize;
  }

  $endtime = $timefield;

  if ($modelist[$RATE] > 0 || $modelist[$INC] > 0) {
    $tag = $timefield - $starttime;
    if ($windowsize > 0) {
      if ($timefield >= $WINDOWend) {
	push(@RATEcount, $WINDOWcount);
	push(@RATEbytes, $WINDOWbytes);
	$WINDOWcount = 0;
	$WINDOWbytes = 0;
	$WINDOWend += $windowsize;
      }
      $WINDOWcount++;
      $WINDOWbytes += $sizeuncomp;
    } else {
      $RATEcount[$tag]++;
      $RATEbytes[$tag] += $sizeuncomp;
    }
  }

  if ($modelist[$SRC] > 0) {
    if ($modelist[$SRC] == 1) {
      $tag = $fields[$inhostfield];
    } elsif ($modelist[$SRC] > 1) {
      $tag = join("\t", $fields[$inhostfield], $fields[$inipfield]);
    }
    $SRCcount{$tag}++;
    $SRCbytes{$tag} += $sizeuncomp;
  }

  if ($modelist[$DIST] > 0) {
    if ($sizeuncomp > $SIZEcachesize) {
      $tag = &getSIZEindex($sizeuncomp);
    } else {
      $tag = $SIZEcache[$sizeuncomp];
    }
    $DISTcount[$tag]++;
    $DISTbytes[$tag] += $sizeuncomp;
  }

  if ($modelist[$USE] > 0 || $modelist[$COMP] > 0) {
    $tag = $fields[$inspoolfield];
    $USEcount[$tag]++;
    $USEbytes[$tag] += $sizecomp;

    {
      use integer;
      $blocks = $sizecomp / 2048;
    }
    $USEsbytes[$tag] += sprintf("%d", ($blocks + 1) * 2048);

    if ($modelist[$COMP] > 0) {
      unless ($USEicount[$tag] > 0) {
	$USEicount[$tag] = 0;
	$USEibytes[$tag] = 0;
	$USEfcount[$tag] = 0;
	$USEfbytes[$tag] = 0;
	$USEucount[$tag] = 0;
	$USEubytes[$tag] = 0;
	$USEccount[$tag] = 0;
	$USEcbytes[$tag] = 0;
	$USEicbytes[$tag] = 0;
	$USEdcount[$tag] = 0;
	$USEdbytes[$tag] = 0;
      }
      $USEicount[$tag] ++;
      $USEibytes[$tag] += $sizeuncomp;
      $USEfbytes[$tag] += $sizecomp;
      if ($sizecomp == 0) {
	$USEdcount[$tag] ++;
	$USEdbytes[$tag] += $sizeuncomp;
      } else {
	$USEfcount[$tag] ++;
	if ($sizeuncomp == $sizecomp) {
	  $USEucount[$tag] ++;
	  $USEubytes[$tag] += $sizeuncomp;
	} else {
	  $USEccount[$tag] ++;
	  $USEicbytes[$tag] += $sizeuncomp;
	  $USEcbytes[$tag] += $sizecomp;
	}
      }
    }
  }
}

&timedata;

# Get last window of data if we have enough
if ($modelist[$RATE] > 0 || $modelist[$INC] > 0) {
  if ($endtime == ($WINDOWend - 1)) {
    push(@RATEcount, $WINDOWcount);
    push(@RATEbytes, $WINDOWbytes);
  }
}

#######################################
## Report

printf("Report from %s to %s, %s\n", $startstr, $endstr, $timestr);

# Incoming Article Rate
if ($modelist[$RATE] > 0) {

  # Specified window size
  if ($windowsize > 0) {
    # Print the headers
    printf("Incoming Article Rate, %d second window\n", $windowsize);
    printf(" %8s %12s %12s %9s %25s\n",
      "Articles", "Bytes", "Start", "End", "Final Second (string)");

    foreach $tag (0 .. $#RATEcount) {
      $count = $RATEcount[$tag];
      $bytes = $RATEbytes[$tag];
      $start = $starttime + ($windowsize * $tag);
      $end = $start + $windowsize - 1;
      $string = localtime($end);

      if ($persecond > 0) {
	printf(" %8.1f %12.1f %12s %9s %25s %6s\n", $count / $windowsize,
	$bytes / $windowsize, $start, $end, $string);
      } else {
	printf(" %8d %12d %12s %9s %25s %6s\n", $count, $bytes, $start,
	$end, $string);
      }
	
    }
  } else {
    # Print the headers
    printf("Incoming Article Rate\n");
    printf("%1s%8s %12s %12s %9s %25s %6s\n",
      "", "Arts", "Bytes", "Second", "Diff", "Second (string)");

    TAG: foreach $tag (0 .. $#RATEcount) {
      $count = $RATEcount[$tag];
      if ($count == 0) {
	next TAG;
      }
      $bytes = $RATEbytes[$tag];
      $end = $starttime + $tag;
      $string = localtime($end);
      if ($tag == 0 || $tag == $#RATEcount) {
	$mark = "*";
      } else {
	$mark = "";
      }

      printf("%1s%8d %12d %12s %9s %25s\n", $mark, $count, $bytes, $end,
	&nicetime($tag - $index), $string);

      $index = $tag;
    }

  }
  printf("\n");
}

# Incoming article source
if ($modelist[$SRC] > 0) {
  # Print the headers
  printf("Incoming Article Source\n");
  if ($modelist[$SRC] == 1) {
    printf("%-40s%-15s %8s %10s\n", "IncomingHostname", "", "Count", "Bytes");
  } elsif ($modelist[$SRC] > 1) {
    printf("%-40s%-15s %8s %10s\n",
      "IncomingHostname", "IP Address", "Count", "Bytes");
  }

  foreach $tag (sort keys %SRCcount) {
    ($host, $ip) = split(/\t/, $tag);
    $count = $SRCcount{$tag};
    $bytes = $SRCbytes{$tag};

    # print results
    if ($persecond > 0) {
      printf("%-40s%-15s %8.3f %10.1f\n", $host, $ip,
        $count / $timediff, $bytes / $timediff);
    } else {
      printf("%-40s%-15s %8d %12s\n", $host, $ip, $count, $bytes);
    }
  }

  # Print totals
  if ($persecond > 0) {
    printf("%-40s%-15s %8.3f %10.1f\n", "TOTAL", "",
      $TOTALcount / $timediff, $TOTALbytes / $timediff);
  } else {
      printf("%-40s%-15s %8d %12s\n", "TOTAL", "", $TOTALcount, $TOTALbytes);
  }

  printf("\n");
}

# Incoming Article Distribution
if ($modelist[$DIST] > 0) {
  # Print the headers
  printf("Incoming Article Distribution by Size\n");
  printf("%3s %8s %10s %10s %12s %12s\n",
    "#", "MaxSize", "Count", "Cum. Count", "Bytes", "Cum. Bytes");

  # Sort article sizes
  @DISTsizes = sort { $a <=> $b } (@DISTsizes);

#  # Initialize counts
#  for $index (0 .. $#SIZE) {
#    $DISTcount[$index] = 0;
#    $DISTbytes[$index] = 0;
#  }
#
  # Start at the first bucket
  $index = 0;
  $max = $SIZE[$index];

#  # Count up all the data
#  foreach $tag (@DISTsizes) {
#
#    # Move onto the correct size group
#    GROUP: while ($tag > $max) {
#      $index++;
#      $max = $SIZE[$index];
#      if ($max == 0) {
#	last GROUP;
#      }
#    }
#
#    # Add data to this bucket
#    $DISTcount[$index]++;
#    $DISTbytes[$index] += $tag;
#  }

  # Display the results
  for $index (1 .. $#SIZE) {
    $DISTtotalcount += $DISTcount[$index];
    $DISTtotalbytes += $DISTbytes[$index];

    # Print numbers as strings to prevent integer overflow
    printf("%3d %8d %10d %10d %12s %12s\n", $index, $SIZE[$index],
      $DISTcount[$index], $DISTtotalcount, $DISTbytes[$index], $DISTtotalbytes);
  }

  # Display totals
  printf("%3s %8s %10d %10s %12s %12s\n", "", "TOTAL", $TOTALcount, "",
    $TOTALbytes, "");
  printf("\n");
}

# Spool Usage
if ($modelist[$USE] > 0) {
  # Print header
  printf("Spool Usage\n");
  printf("%5s %12s %10s %12s %6s %8s %6s %6s\n", "Spool", "Capacity",
    "Articles", "Bytes", "% Cap", "Lifetime", "% Art", "% Byt");

  SPOOL: foreach $index (0 .. $#USEbytes) {
    $bytes = $USEbytes[$index];

    # Check if spool is valid
    if ($bytes == 0) {
      next SPOOL;
    }

    $count = $USEcount[$index];
    $sbytes = $USEsbytes[$index];
    if ($index != 0) {
      $max = &getspoolsize($index);

      # Display
      printf("%5s %12s %10s %12s %5.1f%% %8s %5.1f%% %5.1f%%\n",
	$index, $max, $count, $bytes, ($sbytes / $max) * 100,
	&nicetime(($max / $sbytes) * $timediff),
	($count / $TOTALcount) * 100, ($bytes / $TOTALbytes) * 100);
    } else {
      # Display
      printf("%5s %12s %10s %12s %6s %8s %5.1f%% %5.1f%%\n",
	"none", "", $count, $bytes, "", "",
	($count / $TOTALcount) * 100, ($bytes / $TOTALbytes) * 100);
    }
  }

  # Display total
  printf("%5s %12s %10s %12s\n",
    "TOTAL", "", $TOTALcount, $TOTALbytes);
  printf("\n");

}

# Spool Compression
if ($modelist[$COMP] > 0) {
  printf("Compressed Spool Usage - All Articles\n");
  printf("%5s %8s %11s %11s %9s %8s\n", "", "Article",
    "Initial", "Final", "% Initial", "Comp.");

  printf("%5s %8s %11s %11s %9s %8s\n", "Spool", "Count",
    "Bytes", "Bytes", "Bytes", "Ratio");

  SPOOL: foreach $index (0 .. $#USEibytes) {
    $ibytes = $USEibytes[$index];

    # Check if spool is valid
    if ($ibytes == 0) {
      next SPOOL;
    }

    $fbytes = $USEfbytes[$index];
    $ibytes = $USEibytes[$index];
    $icount = $USEicount[$index];

    if ($index != 0) {
      $max = &getspoolsize($index);

      # Display
      printf("%5s %8s %11s %11s %8.1f%% %7.1f%%\n",
	$index, $icount, $ibytes, $fbytes, ($fbytes/$ibytes) * 100, 
	100 - (($fbytes/$ibytes) * 100));
    } else {
      # Display
      printf("%5s %8s %11s %11s %8.1f%% %8s\n",
	"none", $icount, $ibytes, 0, 0, "-");
    }
  }

  # Display total
  printf("%5s %8s %11s %11s %8.1f%% %7.1f%%\n",
    "TOTAL", $TOTALicount, $TOTALibytes, $TOTALfbytes,
    ($TOTALfbytes/$TOTALibytes) * 100, 100 - (($TOTALfbytes/$TOTALibytes)*100));
  printf("%5s %8s %11s %11s %8.1f%% %7.1f%%\n",
    "NON-D", $TOTALicount - $TOTALdcount, $TOTALibytes - $TOTALdbytes,
    $TOTALfbytes,
    ($TOTALfbytes/($TOTALibytes-$TOTALdbytes)) * 100,
    100 - (($TOTALfbytes/($TOTALibytes-$TOTALdbytes))*100));
  printf("\n");

  if ($TOTALicbytes > 0) {
    printf("Compressed Spool Usage - Compressed Articles Only\n");
    printf("%5s %8s %11s %11s %9s %8s\n", "", "Article",
      "Initial", "Final", "% Initial", "Comp.");

    printf("%5s %8s %11s %11s %9s %8s\n", "Spool", "Count",
      "Bytes", "Bytes", "Bytes", "Ratio");

    SPOOL: foreach $index (0 .. $#USEibytes) {
      $icbytes = $USEicbytes[$index];

      # Check if spool is valid
      if ($icbytes == 0) {
	next SPOOL;
      }

      $cbytes = $USEcbytes[$index];
      $icbytes = $USEicbytes[$index];
      $ccount = $USEccount[$index];

      if ($index != 0) {
	# Display
	printf("%5s %8s %11s %11s %8.1f%% %7.1f%%\n",
	  $index, $ccount, $icbytes, $cbytes, ($cbytes/$icbytes) * 100, 
	  100 - (($cbytes/$icbytes) * 100));
      } else {
      }
    }

    # Display total
    printf("%5s %8s %11s %11s %8.1f%% %7.1f%%\n",
      "TOTAL", $TOTALccount, $TOTALicbytes, $TOTALcbytes,
      ($TOTALcbytes/$TOTALicbytes) * 100,
      100 - (($TOTALcbytes/$TOTALicbytes)*100));
    printf("\n");
  }

}

## Incoming Article Statistics
#if ($modelist[$INC] > 0) {
#  @INCrates = sort { $a <=> $b } values %RATEcount;
#  foreach $rate (@INCrates) {
#    $INCcount[$rate]++;
#  }
#  $INCseconds = $#RATEkeys + 1;
#
#  printf("Incoming Article Statistics\n");
#
#  printf("%12s %12s %12s %12s\n", "", "Seconds", "Articles", "Bytes");
#  printf("%12s %12s %12s %12s\n", "", "", $TOTALcount, $TOTALbytes);
#  printf("%12s %12d %12.2f %12.1f\n", "Active", $INCseconds,
#    $TOTALcount / $INCseconds, $TOTALbytes / $INCseconds);
#  printf("%12s %12d %12.2f %12.1f\n", "Total", $timediff,
#    $TOTALcount / $timediff, $TOTALbytes / $timediff);
#  printf("\n");
#
#  $rate = $INCrates[0];
#  $count = $INCcount[$rate];
#  printf("Low:      %d (%d)\n", $rate, $count);
#
#  $rate = $INCrates[$#INCrates / 2];
#  $count = $INCcount[$rate];
#  printf("Median:   %d (%d)\n", $rate, $count);
#
#  $rate = 0;
#  $count = $INCcount[$index];
#  foreach $index (1 .. $#INCcount) {
#    if ($INCcount[$index] >= $count) {
#      if ($INCcount[$index] == $count) {
#	$rate = join(", ", $rate, $index);
#      } else {
#	$rate = $index;
#      }
#      $count = $INCcount[$index];
#    }
#  }
#  printf("Mode:     %s (%d)\n", $rate, $count);
#
#  $rate = $INCrates[$#INCrates];
#  $count = $INCcount[$rate];
#  printf("High:     %d (%d)\n", $rate, $count);
#}

exit;

#######################################
## Utility subroutines

sub timedata {
  # Do some time parsing
  $startstr = localtime($starttime);
  $endstr = localtime($endtime);
  $timediff = $endtime - $starttime + 1;
  $timestr = &nicetime($timediff);
}

# This utility routine takes a number of seconds and spits out the number
# of days, hours, minutes, seconds it is, only displaying the largest two
# units.  If no time is specified, 0s is returned

sub nicetime {
  my $amount = shift;
  my @vals;
  my $x;
  my ($a, $b);

  use integer;

  # seconds
  $x = $amount % 60;
  if ($x > 0) {
    $x .= "s";
    push(@vals, $x);
  }
  $amount /= 60;

  # minutes
  $x = $amount % 60;
  if ($x > 0) {
    $x .= "m";
    push(@vals, $x);
  }
  $amount /= 60;

  # hours
  $x = $amount % 24;
  if ($x > 0) {
    $x .= "h";
    push(@vals, $x);
  }
  $amount /= 24;

  # days
  $x = $amount;
  if ($x > 0) {
    $x .= "d";
    push(@vals, $x);
  }

  # grab the top two fields
  $a = pop(@vals);
  $b = pop(@vals);
  if ($a ne "") {
    if ($b ne "") {
      $x = join(" ", $a, $b);
    } else {
      $x = $a;
    }
  } else {
    $x = "0s";
  }
}

# Tell me how large the spool I'm asking for is
sub getspoolsize {
  my $spool = shift;
  my $size;

  $size = $SPOOLSIZE[$spool];

  if ($size == 0) {
    $size = $defaultsize;
  }
  $size;
}

# Binary search on SIZE
sub getSIZEindex {
  my $size = shift;

  use integer;

  $left = 0;
  $right = $#SIZE;

  while ($right > ($left + 1)) {
    $mid = ($left + $right) / 2;
    if ($size <= $SIZE[$mid - 1]) {
      $right = $mid;
    } else {
      $left = $mid;
    }
  }

  $left;
}
